"""
CLI commands for docker actions
"""


import json
import tempfile
from os import makedirs, path

import click

from brewblox_dev import utils

WORKDIR = path.expanduser('~/.cache/brewblox-dev/docker')

PYTHON_TAGS = [
    '3.8', '3.8-slim',
]
NODE_TAGS = [
    '12', '12-alpine',
]

AMD_ARM_REPOS = [
    'brewblox/brewblox-ctl-lib',
    'brewblox/brewblox-devcon-spark',
    'brewblox/brewblox-history',
    'brewblox/brewblox-mdns',
    'brewblox/brewblox-ui',
    'brewblox/firmware-flasher',
    'brewblox/brewblox-emitter',
    'brewblox/brewblox-plaato',
    'brewblox/brewblox-automation',
]
AMD_REPOS = [
    'brewblox/firmware-simulator',
]


@click.group()
def cli():
    """Collection group"""


def enable_experimental():
    cfg = '/etc/docker/daemon.json'
    try:
        with open(cfg, 'r') as f:
            content = f.read()
            config = json.loads(content or '{}')
    except FileNotFoundError:  # pragma: no cover
        config = {}

    if config.get('experimental'):
        return

    h, temp = tempfile.mkstemp()

    with open(temp, 'w') as f:
        config['experimental'] = True
        json.dump(config, f, indent=4)

    utils.run(f'sudo mv -f {temp} {cfg}')
    utils.run('sudo systemctl restart docker')


def install_qemu():
    utils.run('sudo apt update')
    utils.run('sudo apt install -y qemu qemu-user-static qemu-user binfmt-support')
    utils.run(f'cp $(which qemu-arm-static) {WORKDIR}/')


def build_python_images():
    for tag in PYTHON_TAGS:
        with open(f'{WORKDIR}/Dockerfile', 'w') as f:
            f.write('\n'.join([
                f'FROM arm32v7/python:{tag}',
                'COPY ./qemu-arm-static /usr/bin/qemu-arm-static',
            ]))
        utils.run(f'docker pull --platform=linux/arm arm32v7/python:{tag}')
        utils.run(f'docker build --platform=linux/arm --no-cache -t brewblox/rpi-python:{tag} {WORKDIR}')
        utils.run(f'docker push brewblox/rpi-python:{tag}')


def build_node_images():
    for tag in NODE_TAGS:
        with open(f'{WORKDIR}/Dockerfile', 'w') as f:
            f.write('\n'.join([
                f'FROM arm32v7/node:{tag}',
                'COPY ./qemu-arm-static /usr/bin/qemu-arm-static',
            ]))
        utils.run(f'docker pull --platform=linux/arm arm32v7/node:{tag}')
        utils.run(f'docker build --platform=linux/arm --no-cache -t brewblox/rpi-node:{tag} {WORKDIR}')
        utils.run(f'docker push brewblox/rpi-node:{tag}')

    # Alpine is an exception, as it runs on armv6, and must install packages
    with open(f'{WORKDIR}/Dockerfile', 'w') as f:
        f.write('\n'.join([
            'FROM arm32v6/alpine',
            'COPY ./qemu-arm-static /usr/bin/qemu-arm-static',
            'RUN apk add --no-cache nodejs npm',
        ]))

    utils.run('docker pull --platform=linux/arm arm32v6/alpine')
    utils.run(f'docker build --platform=linux/arm --no-cache -t brewblox/rpi-node:10-alpine {WORKDIR}')
    utils.run('docker push brewblox/rpi-node:10-alpine')


@cli.command()
def docker_info():
    print('Stash directory:', WORKDIR)
    print('Python tags', *PYTHON_TAGS, sep='\n\t')
    print('Node tags', *NODE_TAGS, sep='\n\t')
    print('AMD+ARM repositories:', *AMD_ARM_REPOS, sep='\n\t')
    print('AMD repositories:', *AMD_REPOS, sep='\n\t')


@cli.command()
@click.option('--python', is_flag=True, help='Build and push Python images')
@click.option('--node', is_flag=True, help='Build and push Node images')
def docker_images(python, node):
    """Create base images used by various BrewBlox projects"""
    makedirs(WORKDIR, exist_ok=True)
    enable_experimental()
    install_qemu()

    if python:
        build_python_images()

    if node:
        build_node_images()


@cli.command()
def release_stable():
    """Push newest-tag -> stable for all managed repositories"""
    for repo in AMD_ARM_REPOS + AMD_REPOS:
        utils.run(f'docker pull {repo}:newest-tag')

    for repo in AMD_ARM_REPOS:
        utils.run(f'docker pull {repo}:rpi-newest-tag')

    # start pushing after all pulls are ok

    for repo in AMD_ARM_REPOS + AMD_REPOS:
        utils.run(f'docker tag {repo}:newest-tag {repo}:stable')
        utils.run(f'docker push {repo}:stable')

    for repo in AMD_ARM_REPOS:
        utils.run(f'docker tag {repo}:rpi-newest-tag {repo}:rpi-stable')
        utils.run(f'docker push {repo}:rpi-stable')



"""
Commands for git operations
"""

from contextlib import suppress
from distutils.version import StrictVersion
from os import makedirs, path
from shutil import which
from subprocess import STDOUT, CalledProcessError, check_call, check_output

import click

from brewblox_dev import utils

WORKDIR = path.expanduser('~/.cache/brewblox-dev/git')
REPOS = [
    'brewblox-documentation',
    'brewblox-devcon-spark',
    'brewblox-history',
    'brewblox-ui',
    'brewblox-ctl-lib',
    'brewblox-firmware',
    'brewblox-plaato',
    'brewblox-tilt',
    'brewblox-hass',
    'brewblox-images',
    'node-red-contrib-brewblox',
]


@click.group()
def cli():
    """Command collector"""


def create_repos():
    makedirs(WORKDIR, exist_ok=True)
    [
        check_output(
            f'git clone --no-checkout https://github.com/BrewBlox/{repo}.git', shell=True, cwd=WORKDIR)
        for repo in REPOS
        if not path.exists(f'{WORKDIR}/{repo}/.git')
    ]


def install_hub():
    check_output('sudo apt-get install -y hub', shell=True)


def prepare():
    create_repos()
    if not which('hub') and utils.confirm('hub cli not found - do you want to install it?'):
        install_hub()


@cli.command()
def git_info():
    print('Stash directory:', WORKDIR)
    print('Github repositories:', *REPOS, sep='\n\t')
    check_call('hub --version', shell=True)


@cli.command()
def delta():
    """Show commit delta for all managed repositories"""
    prepare()

    headers = ['repository'.ljust(25), 'develop >', 'edge >', 'tag']
    print(*headers)
    # will include separators added by print()
    print('-' * len(' '.join(headers)))
    for repo in REPOS:
        check_output('git fetch --tags --quiet',
                     shell=True,
                     cwd=f'{WORKDIR}/{repo}')
        dev_edge = check_output(
            'git rev-list --count origin/edge..origin/develop',
            shell=True,
            cwd=f'{WORKDIR}/{repo}').decode().rstrip()
        edge_tag = check_output(
            'git rev-list --count $(git rev-list --tags --max-count=1)..origin/edge',
            shell=True,
            cwd=f'{WORKDIR}/{repo}').decode().rstrip()
        vals = [repo, dev_edge, edge_tag, '-']
        print(*[v.ljust(len(headers[idx])) for idx, v in enumerate(vals)])


@cli.command()
def compare():
    """Show GitHub comparison URLs for all managed repositories"""
    for repo in REPOS:
        print(f'https://github.com/BrewBlox/{repo}/compare/edge...develop')


@cli.command()
def release_edge():
    """Create develop -> edge PRs for all managed repositories"""
    prepare()

    for repo in REPOS:
        if not utils.confirm(f'Do you want to create a develop -> edge PR for {repo}?'):
            continue

        with suppress(CalledProcessError):
            check_call(
                'hub pull-request -b edge -h develop -m "edge release"',
                shell=True,
                cwd=f'{WORKDIR}/{repo}')


def bumped_version(current_version: str, bump_type: str) -> str:
    major, minor, patch = StrictVersion(current_version).version

    return {
        'major': lambda: f'{major + 1}.{0}.{0}',
        'minor': lambda: f'{major}.{minor + 1}.{0}',
        'patch': lambda: f'{major}.{minor}.{patch + 1}',
    }[bump_type]()


@cli.command()
@click.argument('increment', required=True, type=click.Choice(['major', 'minor', 'patch']))
@click.option('--init', is_flag=True)
def bump(increment, init):
    """Increases git tag containing semantic version"""
    if init:
        current_version = '0.0.0'
    else:
        # Get all version-formatted tags, but use the latest
        current_version = check_output(
            r'git tag -l *.*.* --contains $(git rev-list --tags --max-count=1)',
            shell=True
        ).decode().rstrip().split('\n')[-1]

    new_version = bumped_version(current_version, increment)

    print(f'Bumping "{increment}" version: {current_version} ==> {new_version}')

    if utils.confirm('Do you want to tag the current commit with that version?'):
        check_output(f'git tag -a {new_version} -m "Version {new_version}"', shell=True)

        print('Latest tags:')
        print(check_output('git tag --sort=-version:refname -n1 | head -n5', shell=True).decode().rstrip())

    else:
        print('Aborted. No tags were added!')
        return

    if utils.confirm('Do you want to push this tag?'):
        check_call('git push --tags', shell=True, stderr=STDOUT)



"""
Entrypoint for the brewblox-dev tool
"""

import click
from dotenv import find_dotenv, load_dotenv

from brewblox_dev import docker, localbuild, repository

cli = click.CommandCollection(
    sources=[
        localbuild.cli,
        repository.cli,
        docker.cli,
    ])


def main():  # pragma: no cover
    load_dotenv(find_dotenv(usecwd=True))
    cli()


if __name__ == '__main__':
    main()



#! /usr/bin/python3


import re
from glob import glob
from os import getenv, remove
from subprocess import check_output

import click

from brewblox_dev import utils


@click.group()
def cli():
    """Command collector"""


@cli.command()
@click.option('-a', '--arch', multiple=True, type=click.Choice(['amd', 'arm']),
              help='Build target. Can be repeated. Default: amd')
@click.option('-r', '--repo', default=lambda: getenv('DOCKER_REPO'),
              help='Docker repository name.')
@click.option('-c', '--context', default='docker',
              help='Docker build context.')
@click.option('-f', '--file', default='Dockerfile',
              help='Filename inside build context.')
@click.option('-t', '--tag', multiple=True,
              help='Additional tag. the "local" tag is always built.')
@click.option('--branch-tag', is_flag=True,
              help='Use sanitized branch name as tag. ARM automatically gets prefixed with "rpi-".')
@click.option('--setup/--no-setup', default=True,
              help='Perform Python-related setup steps.')
@click.option('--pull', is_flag=True,
              help='Pull base images.')
@click.option('--push', is_flag=True,
              help='Push all tags except "local" to Docker Hub.')
def localbuild(arch, repo, context, file, tag, branch_tag, setup, pull, push):
    """Build docker containers"""
    arch = list(arch or ['amd'])
    tags = list(tag)

    if branch_tag:
        tags.append(getenv('TRAVIS_BRANCH')  # Travis
                    or getenv('Build.SourceBranchName')  # Azure Pipelines
                    or check_output('git rev-parse --abbrev-ref HEAD'.split()).decode().rstrip())

    tags = [re.sub('[/_:]', '-', t) for t in tags]

    if setup:
        for f in glob('dist/*'):
            remove(f)

        utils.run('python setup.py sdist')
        utils.run(f'pipenv lock --requirements > {context}/requirements.txt')
        utils.distcopy('dist/', [f'{context}/dist/'])

    # single command
    utils.run(f'cd {context}'
              ' && if [ -f ./localbuild.sh ]; then bash ./localbuild.sh; fi')

    for a in arch:
        prefix = ''
        commands = []
        build_args = []
        build_tags = [
            'local',
            *tags
        ]

        if a == 'arm':
            prefix = 'rpi-'
            commands += [
                'docker run --rm --privileged multiarch/qemu-user-static:register --reset',
            ]

        if pull:
            build_args.append('--pull')

        build_args += [
            '--build-arg service_info="$(git describe) @ $(date)"',
            '--no-cache',
            ' '.join([f'--tag {repo}:{prefix}{t}' for t in build_tags]),
            f'--file {context}/{a}/{file}',
            context,
        ]

        commands.append(f'docker build {" ".join(build_args)}')
        utils.run(' && '.join(commands))

        if push:
            # We're skipping the local tag
            [utils.run(f'docker push {repo}:{prefix}{t}') for t in tags]



"""
Utility functions
"""

import pathlib
from distutils.util import strtobool
from glob import glob
from os import remove
from shutil import copy
from subprocess import STDOUT, check_call


def run(cmd: str):
    print(cmd)
    check_call(cmd, shell=True, stderr=STDOUT)


def confirm(question):
    print('{} [Y/n]'.format(question))
    while True:
        try:
            return bool(strtobool(input().lower() or 'y'))
        except ValueError:
            print('Please respond with \'y(es)\' or \'n(o)\'.')


def distcopy(source, destinations):

    for dest in destinations:
        if not dest or dest == '/':
            raise AssertionError(f'Removing "{dest}/*" is a BAD idea, hmkay?')

        pathlib.Path(dest).mkdir(parents=True, exist_ok=True)

        for f in glob(dest + '/*'):
            remove(f)

        for f in glob(source + '/*'):
            copy(f, dest + '/')
            print(f'{f} ==> {dest}')
